#include "Python.h"

#define TYPE char
#define OBJ_NAME xobject
#define FN_SPLIT_NEW xsplit_new
#define FN_XITER_NEXT xiter_next
#define FN_STRING_BUILD PyString_FromStringAndSize
#define TYPE_XITER XIter_Type
#define PyTYPE_AS_TYPE PyString_AS_STRING
#define PyTYPE_GET_SIZE PyString_GET_SIZE
#define PyTYPE_CHECK PyString_CheckExact

typedef struct {
	PyObject_HEAD
	TYPE *src;
	TYPE *srcEnd;
	TYPE *index;
	TYPE *to;
	TYPE *splitter;
	TYPE *splitterEnd;
} OBJ_NAME;

PyAPI_DATA(PyTypeObject) TYPE_XITER;
static PyObject *FN_SPLIT_NEW(PyTypeObject *type, PyObject *arg, PyObject *kw);
static PyObject *FN_XITER_NEXT(OBJ_NAME *r);

PyTypeObject TYPE_XITER = {
	PyObject_HEAD_INIT(&PyType_Type)
	0, "xiter",
	sizeof(OBJ_NAME),0,
	(destructor)PyObject_Del,
	0,0,0,0,0,0,0,0,0,0,0,
	PyObject_GenericGetAttr,0,0,
	Py_TPFLAGS_DEFAULT,0,0,0,0,0,
	PyObject_SelfIter,
	(iternextfunc)FN_XITER_NEXT,
	0,0,0,0,0,0,0,0,0,0,
	FN_SPLIT_NEW,
};

static PyObject *
FN_SPLIT_NEW(PyTypeObject *type, PyObject *arg, PyObject *kw) {
	PyObject *src, *splitter;

	if (!PyArg_UnpackTuple(arg, "", 1, 2, &src, &splitter))
		return NULL;

	if (!PyTYPE_CHECK(src)) {
		return NULL;
	}

	OBJ_NAME *obj = PyObject_New(OBJ_NAME, &TYPE_XITER);
	if (obj == NULL) return NULL;

	obj->src = PyTYPE_AS_TYPE(src);
	obj->srcEnd = obj->src + PyTYPE_GET_SIZE(src);
	obj->index = obj->to = obj->src;

	if (splitter != NULL) {
		if (!PyTYPE_CHECK(splitter)) {
			return NULL;
		}
		obj->splitter = PyTYPE_AS_TYPE(splitter);
		obj->splitterEnd = obj->splitter + PyTYPE_GET_SIZE(splitter);
	}

	return (PyObject *) obj;
}

static PyObject *
FN_XITER_NEXT(OBJ_NAME *r) {
	Py_ssize_t splitterSize = r->splitterEnd - r->splitter;

	TYPE *s = r->index, *start = s, *ss, *sEnd = r->srcEnd;
	TYPE *splitter = r->splitter, *sp, *spEnd = r->splitterEnd;
	while (s < sEnd) {
		if (*s == *splitter) {
			if (splitterSize > 1) {
				sp = splitter, ss = s;
				while(*++sp == *++ss);
				if (sp >= spEnd) {
					r->to = s;
					r->index = s + splitterSize;
					s+= splitterSize;
					return FN_STRING_BUILD(start, r->to - start);
				}
			} else {
				r->to = s;
				r->index = s + splitterSize;
				return FN_STRING_BUILD(start, r->to - start);
			}
		}
		s++;
	}
	if (r->to < r->srcEnd) {
		r->to = s;
		r->index = r->srcEnd;
		return FN_STRING_BUILD(start, r->to - start);
	}
	return NULL;
}

////////////////////////////////////////////////////////////

#undef TYPE
#undef OBJ_NAME
#undef FN_SPLIT_NEW
#undef FN_XITER_NEXT
#undef FN_STRING_BUILD
#undef TYPE_XITER
#undef PyTYPE_AS_TYPE
#undef PyTYPE_GET_SIZE
#undef PyTYPE_CHECK

#define TYPE Py_UNICODE
#define OBJ_NAME xobject_u
#define FN_SPLIT_NEW xsplit_new_u
#define FN_XITER_NEXT xiter_next_u
#define FN_STRING_BUILD PyUnicode_FromUnicode
#define TYPE_XITER XIter_Type_U
#define PyTYPE_AS_TYPE PyUnicode_AS_UNICODE
#define PyTYPE_GET_SIZE PyUnicode_GET_SIZE
#define PyTYPE_CHECK PyUnicode_CheckExact

typedef struct {
	PyObject_HEAD
	TYPE *src;
	TYPE *srcEnd;
	TYPE *index;
	TYPE *to;
	TYPE *splitter;
	TYPE *splitterEnd;
} OBJ_NAME;

PyAPI_DATA(PyTypeObject) TYPE_XITER;
static PyObject *FN_SPLIT_NEW(PyTypeObject *type, PyObject *arg, PyObject *kw);
static PyObject *FN_XITER_NEXT(OBJ_NAME *r);

PyTypeObject TYPE_XITER = {
	PyObject_HEAD_INIT(&PyType_Type)
	0, "xiter",
	sizeof(OBJ_NAME),0,
	(destructor)PyObject_Del,
	0,0,0,0,0,0,0,0,0,0,0,
	PyObject_GenericGetAttr,0,0,
	Py_TPFLAGS_DEFAULT,0,0,0,0,0,
	PyObject_SelfIter,
	(iternextfunc)FN_XITER_NEXT,
	0,0,0,0,0,0,0,0,0,0,
	FN_SPLIT_NEW,
};

static PyObject *
FN_SPLIT_NEW(PyTypeObject *type, PyObject *arg, PyObject *kw) {
	PyObject *src, *splitter;

	if (!PyArg_UnpackTuple(arg, "", 1, 2, &src, &splitter))
		return NULL;

	if (!PyTYPE_CHECK(src)) {
		return NULL;
	}

	OBJ_NAME *obj = PyObject_New(OBJ_NAME, &TYPE_XITER);
	if (obj == NULL) return NULL;

	obj->src = PyTYPE_AS_TYPE(src);
	obj->srcEnd = obj->src + PyTYPE_GET_SIZE(src);
	obj->index = obj->to = obj->src;

	if (splitter != NULL) {
		if (!PyTYPE_CHECK(splitter)) {
			return NULL;
		}
		obj->splitter = PyTYPE_AS_TYPE(splitter);
		obj->splitterEnd = obj->splitter + PyTYPE_GET_SIZE(splitter);
	}

	return (PyObject *) obj;
}

static PyObject *
FN_XITER_NEXT(OBJ_NAME *r) {
	Py_ssize_t splitterSize = r->splitterEnd - r->splitter;

	TYPE *s = r->index, *start = s, *ss, *sEnd = r->srcEnd;
	TYPE *splitter = r->splitter, *sp, *spEnd = r->splitterEnd;
	while (s < sEnd) {
		if (*s == *splitter) {
			if (splitterSize > 1) {
				sp = splitter, ss = s;
				while(*++sp == *++ss);
				if (sp >= spEnd) {
					r->to = s;
					r->index = s + splitterSize;
					s+= splitterSize;
					return FN_STRING_BUILD(start, r->to - start);
				}
			} else {
				r->to = s;
				r->index = s + splitterSize;
				return FN_STRING_BUILD(start, r->to - start);
			}
		}
		s++;
	}
	if (r->to < r->srcEnd) {
		r->to = s;
		r->index = r->srcEnd;
		return FN_STRING_BUILD(start, r->to - start);
	}
	return NULL;
}

static PyMethodDef exports[] = {
	{"split", (PyCFunction)xsplit_new, METH_VARARGS, "split iter for str"},
	{"split_u", (PyCFunction)xsplit_new_u, METH_VARARGS, "split iter for unicode"},
	{NULL, NULL}
};

PyMODINIT_FUNC
initxiter(void) {
	Py_InitModule("xiter", exports);
}
